package de.korap.services;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.io.PrintStream;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.time.LocalDateTime;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import de.korap.services.utils.Utils;

/* 0.1 : 17.06.20/FB 
 * 0.2 : 25.10.21/FB
 *       loging of the API Requests.
 *       loading constants from GlemmServices.properties.
 *       correcting calling of lemma + prop1/prop2/spec.
 *       handling of lemmata with 'ß'.
 * 
 * General comments to the encoding of derivations from lemma to wfs and how they are encoded
 * in the Derby database:
 * 
 * GLEMM encoding:  
 *   _ _ : +Flex       Öle -> Öl      
 *   + _ : +Comp       Öllampe -> Öl or Lampe (compound).
 *   _ + : +other      Antiturnschuh -> Anti- + Turnschuh (no compound for 'Turnschuh', only prefix).
 *   + + : +Comp+other Antiturnschuh -> Anti- + Turn + Schuh (compound for 'Schuh' + Prefix).
 * C2 lemmatization options:
 *   Flex  (identical to Glemm)
 *   Comp  (identical to Glemm, includes all Flexions)
 *   Other (identical to Glemm, includes all Flexions).
 *   Spec  (only C2: includes all wfs with hiphen and/or some non-alphabetic ISO8859-1 chars)
 *         E.g. Öl-Exporte, "Marken"-Öle.
 * Derby Encoding:
 *  prop1 	prop2	spec   	meaning in C2:
 *  0		0		-		Flex
 *  1		0		-		Comp
 *  0		1		-		Other (prefix)
 *  1		1		-		Comp + Other
 *  -		-		0		-Spec
 *  -		-		1		+Spec
 * 
 * Examples:
 * &Schuh			: Schuh, Schuhe, ... only flexions
 * &Schuh+comp		: Schuhe, Turnschuhe... flexions and compounds.
 * &Schuh+other		: Schuhe, Antischuhe... flexions and prefixed wfs (no compounds).
 * &Schuh+comp+other: Schuhe, Antischuhe, Turnschuhe, Antiturnschuhe... flexions and compounds and 
 * 					  prefixed single wfs (Antischuhe) and prefixed compounds.
 * &Schuh+spec		: "Anti"-Schuhe, 68ger-Schuhe... may be combined with every options: adds
 *                    wfs with hyphen and/or non-alphabetic chars.
 * 
 * Conversion rules:
 * 
 *  &schließen -> internal search for 'schliessen',
 *  because Glemm outputs 'ss' for 'ß'.
 *  
 * Character Encoding:
 * 
 * - When importing wfs to the Derby Database with doCurlGlemmServices/indexLemmata:
 * 
 *   INPUT: wordform list in [UTF-8] : UTF-8 is then converted to Lat1 for Glemm,
 *          the output of Glemm [Lat1] is returned to the GlemmServices without reconversion
 *          to UTF8 and imported into Derby Database unchanged.
 *		
 */

@Path("/index")

public class GlemmServices {
	
	final private String versionDate	= "05.11.21"; 
	final private String version		= "0.2";
	
	final private String fnameProps		= "/WEB-INF/GlemmServices.properties";
	
	// the following constants are read from GlemmServices.properties (these are the default values):
	private String GS_WorkingPath 		= "/home/bodmer/KorAP/GlemmServices/Tests";
	private String fnameWfs 			= GS_WorkingPath + "/gl_wfs_in.txt";
	private String fnameOut 			= GS_WorkingPath + "/GlemmServices.log";
	private String fnameErr 			= GS_WorkingPath + "/GlemmServices.err";
	private String fnameScript 			= GS_WorkingPath + "/doGlemm";
	private String fnameLemata			= GS_WorkingPath + "/tmp/lem3.utf8";

	@Context
	private HttpServletRequest httpServletRequest;
	@Context
	private ServletContext servletContext;
	
	final private int	ACC_FORMAT_TEXT 	= 1;
	final private int	ACC_FORMAT_JSON		= 2;
	final private int	ACC_FORMAT_HTML		= 3;
	
	public static PrintStream fout	= null;
	public static PrintStream ferr	= null;		
		
	/* 
 	 * 
	 * - after loadProperties, sets the class variable with loaded property values.
	 * 
	 * 15.10.21/FB
	 */
	
	private void set_ConstantsFromProperties(Properties props)
	
	{
	GS_WorkingPath 	= Utils.getConstantFromProperties(props, "GS_WorkingPath", "/home/bodmer/KorAP/GlemmServices/Tests");
	fnameWfs		= Utils.getConstantFromProperties(props, "fnameWfs", 	   "gl_wfs_in.txt");
	fnameWfs		= GS_WorkingPath + "/" + fnameWfs;
	fnameOut		= Utils.getConstantFromProperties(props, "fnameOut", 	   "GlemmServices.log");
	fnameOut		= GS_WorkingPath + "/" + fnameOut;
	fnameErr		= Utils.getConstantFromProperties(props, "fnameErr", 	   "GlemmServices.err");
	fnameErr		= GS_WorkingPath + "/" + fnameErr;
	fnameScript		= Utils.getConstantFromProperties(props, "fnameScript",	   "doGlemm");
	fnameScript		= GS_WorkingPath + "/" + fnameScript;
	fnameLemata		= Utils.getConstantFromProperties(props, "fnameLemata",    "tmp/lem3");
	fnameLemata		= GS_WorkingPath + "/" + fnameLemata;
	} // set_ConstantsFromProperties
	
	/*
	 *  log_Request:
	 *  
	 *  - write info about API Request.
	 *  Parameters:
	 *  requestPath	: path of webapp addressed by the request.
	 *  func		: name of the function which implements the request.
	 *  
	 *  19.10.21/FB
	 */
	
	private void log_Request(String requestPath, String func)
	
	{
	// transform "dateTtime" -> "date time":
	
	fout.printf("###\n");
	fout.printf("### %s: %s : vers='%s', vdate='%s' request='%s'.\n", func, requestPath, version, versionDate, 
						LocalDateTime.now().toString().replace("T",  " "));
	fout.printf("###\n");
	} // log_Request
		
	/* formatDbName
	 * 
	 * - gets name of Database used by Derby from props 
	 *   and format it to HTML.
	 *  
	 * 19.10.21/FB
	 */
	
	private String formatDbName(Properties props)
	
	{
	String
		dbName;
	String
		dbPath = Utils.getConstantFromProperties(props, "fnameDB", "/home/bodmer/KorAP/GlemmServices/Dbs/DerbyTestDB");
	int
		j = dbPath.lastIndexOf('/');
	
	if( j > 0 )
		dbName = dbPath.substring(j+1);
	else
		dbName = dbPath;
	
	return dbName; // "<p><b>Database: " + dbName + "</b></p>\n";
	} // formatDbName

	/*
	 * closeLogStreams:
	 * 
	 * - close the fout/ferr log streams.
	 * 
	 * 26.10.21/FB
	 */
	
	private void closeLogStreams()
	
	{
	if( fout != null)
		fout.close();
	fout = null;
	
	if( ferr != null )
		ferr.close();
	ferr = null;
	} // closeLogStreams
	
	/* init_and_log_Request
	 * 
	 * - load GlemmServices.properties, sets GlemmServices constants
	 *   and logs the request.
	 * - load properties and set constants before logging anything, because
	 *   logging will be done into the last stdout/stderr opened in glassfish's JVM,
	 *   e.g. in the log of another Service.
	 * 19.10.21/FB
	 */
	
	private Properties init_and_log_Request(String func, String requestPath)
	
	{
	String
		realPath = servletContext.getRealPath(fnameProps);
	
	// load GlemmServices properties and log them (if last param = true):
	Properties
		props = Utils.loadProperties(servletContext.getResourceAsStream(fnameProps), fnameProps, realPath, false);
	
	set_ConstantsFromProperties(props);
	
	// open fout and ferr as replacement for stdout/stderr for use in GlemmServices.
	// do not use or redirect System.out and System.err, as they will be redirected
	// by other Services in the same glassfish JVM, e.g. by the WformServices.
	fout = Utils.setLogStream(fnameOut);
	ferr = Utils.setLogStream(fnameErr);
	
	//Utils.setStdFileStreams(fnameErr, fnameOut);
	
	log_Request(requestPath, func);
	
	return props;
	} // init_and_log_Request

	/* * * * * * * * * * * * * * * * * * 
	 * 
	 * doGET_status() on /index/status.
	 * 
	 * GET status of Glemm Services
	 * 
	 * 17.06.20/FB
	 * 
	 * * * * * * * * * * * * * * * * * *
	 */
	
	@Path("status")
	@GET
	@Produces("text/html")
	
	public String doGET_status() 
	
	{
	final String func = "doGET_status";
	String
		status;
	Properties
		props;
	boolean
		bDerbyServerConnected = false;
	
	props = init_and_log_Request(func, "/index/status/");
	
	bDerbyServerConnected = GlemmDB.checkDerbyConnection() == 0 ? true : false;
	
	// httpServletRequest.setCharacterEncoding(env);
	
	status = String.format(	"<html>\n" +
						   	" <h3>Glemm Services</h3>\n" +
						   	" <p><b>Version</b> : %s</p>" +
							" <p><b>Version date</b> : %s</p>" +
							" <p><b>Services status</b>  : ready.</p>" +
							" <p><b>Derby Server status</b> : %s.</p>" +
							" <p><b>Request time</b>: %s</p>" +
							" <p><b>properties</b>: %s </p>" +
							" <p><b>database</b>: %s </p>" +
							"</html>",
							version,
							versionDate,
							bDerbyServerConnected ? "connected" : "not available",
							LocalDateTime.now().toString().replace("T",  " "),
							servletContext.getResourceAsStream(fnameProps) == null ? "not found" :
								props.isEmpty() ? "empty" : "loaded",
							formatDbName(props)
							);

	fout.printf("Debug: %s: done.\n", func);
	closeLogStreams();
	return status;	
		
	} // doGET_status
	
	/* ***********************************************************
	 * 
	 * doPOST_importLemmaData:
	 * 
	 * POST: /index/add/{databasename}: text/plain
	 * 
	 * Parameters:
	 * list		 	: the list of wfs sent in the body of the POST request.
	 * dbpathEncoded: the location of the database where the lemma+wfs are to be stored.
	 *             	  encoded as ".home.user.KorAP....", must be converted to
	 *            	  "/home/user/KorAP/..."
	 *         		  IMPORTANT: do not use a '.' for encoding anything else than a '/'.
	 * Notes     :
	 * - adds a list of new wordforms to the GlemmService's Database located at 'databasename'.
	 * - The wfs are lemmatized and the morphological data stored into a Derby database.
	 * Input	 : wordform data expected in [UTF-8].
	 * 25.11.19/FB
	 * 04.11.21/FB
	 * - submitted database path.
	 */
	
	@Path("add/{databasepath}") 
	@POST 
	@Consumes("text/plain") 
	@Produces("text/html")
	 
	public String doPOST_importLemmaData
							(
							String 										  list,
							@PathParam(value="databasepath") final String dbpathEncoded
							) 
	 
	{
	final String
		func = "doPOST_importLemmaData";
	String[] 
		arList = list.split("\n");
	String
		dbpath;
	FileSystem 
		fs = FileSystems.getDefault();
	java.nio.file.Path
		p = null;
	OutputStream 
		os;
	boolean
		bDone = false;
	String
		msg;
	int
		nErrors = 0;
	Properties
		props;

	props = init_and_log_Request(func, "/index/add/{dbpath}");
	
	dbpath = dbpathEncoded.replace('.', '/');
	
	fout.printf("Info : %s: selected database path = '%s'.\n", func, dbpath);
	fout.printf("Info : %s: working directory = '%s'.\n", func, GS_WorkingPath);
	
	// set file and paths after GS_WorkingPath has been initialised:
	p = fs.getPath(fnameWfs);
	
	// Write submitted list of wordforms to file:
	try
		{
		os = Files.newOutputStream(p);
		os.write(list.getBytes(StandardCharsets.ISO_8859_1));
	    os.close();
	    bDone = true;
	    fout.printf("Debug: %s: file '%s' beschrieben.\n", func, fnameWfs);
	    }
	catch(IOException e)
		{
		e.printStackTrace(ferr);
		}
    
	if( bDone )
		msg = "<p>List written to file: " + p.toString() + "</p>";
	else
		msg = "<p><b>File Error: </b> cannot write to " + p.toString() + "!</p>";
	
	/*
	 *  execute Glemm shell script:
	 */
	String
		msgGlemmScript, msgImport;
	int 
		r = Utils.executeShellScript(fnameScript, fnameWfs, fout, ferr);
	
	if( r == 0 )
		msgGlemmScript = "<p><b>Glemm Script:</b> Ausführung OK.</p>";
	else
		msgGlemmScript = "<p><b>Glemm Script:</b> Fehler aufgetreten!";
	
	// Build a reponse and return it:
	if( r != 0 )
		return 
		   "<html>" +
		   "<h1>GlemmServices: POST on /index/add: add List of wordforms.</h1>" +
		   "<p>Version vom " + versionDate + "</p>" +
		   "<p>Consumes = text/plain</p>" +
		   "<p>List length =" + arList.length +
		   msg + msgGlemmScript +
		   "</html>"; 
	
	
	// Read lemmatized data and store them into database:
	nErrors = GlemmDB.importGlemmLemmata(fnameLemata, props, dbpath);
	
	fout.printf("Debug: %s: importGlemmLemmata returns r=%d.\n", func, r);
	
	if( nErrors == 0 )
		msgImport = "<p><b>Import:</b> " + fnameLemata + ": OK.</p>";
	else
		msgImport = "<p><b>Import:</b> " + fnameLemata + ": " + nErrors + " errors!</p>";
	
	fout.printf("Debug: %s: ending time: %s.\n", func, LocalDateTime.now().toString().replace("T",  " "));

	closeLogStreams();
	
	// Return Infos:
	return 
			   "<html>" +
			   "<h1>GlemmServices: POST on /index/add: add List of wordforms.</h1>" +
			   "<p>Version vom " + versionDate + "</p>" +
			   "<p>Consumes = text/plain</p>" +
			   "<p>List length =" + arList.length +
			   msg + msgGlemmScript + msgImport +
			   "</html>";
	
	} // doPOST_importLemmaData

	/* doGET_viewDbInfo:
	 * 
	 * GET /index/info/: text/plain > text/html. 
	 * Returns general infos about the current content of the
	 * Glemm Lemma Database.
	 * 10.01.20/FB
	 */
	@GET @Path("info") @Produces("text/html")
	 
	public String doGET_viewDbInfo() 
	 
	{
	final String
		func = "doGET_viewDbInfo";
	int
		i, r;
	String
		returnContent;
	final String
		returnHeader = 	"<html><head>\n" + 
						" <meta charset=\"utf-8\">\n" + 
						"</head>\n" + 
						"<h1>GlemmServices: GET on /index/info</h1>\n" +
						"<h4>(vers = " + versionDate + ")</h4>\n" + 
						"<h2>Current state of the Database</h2>\n";
	final String
		returnTrailer = "</html>",
		alri          = "align='right'";
	StringBuilder
		table = new StringBuilder();
	String
		msg = null;
	GlemmDBInfo
		dbInfo = null;
	Properties
		props;
	
	props = init_and_log_Request(func, "/index/info/");
	
	dbInfo = GlemmDB.readDbInfos(props);
	if( dbInfo == null )
		table.append("<p><b>Fehler:</b> Inhalt der Datebank kann nicht gelesen werden!</p>");
	else
		{
		table.append("<table border=\"1\" cellpadding=\"2\" cellspacing=\"2\">\n");
	
		for(i=0; i<dbInfo.counts.length; i++)
			{
			msg = String.format(Locale.GERMAN, " <tr><td><b>Anzahl %s</b></td><td %s>%,d</td></tr>\n", dbInfo.countLabels[i], alri, dbInfo.counts[i]);
			table.append(msg);
			}
		table.append("</table>");
		}

	closeLogStreams();
	
	final String
		dbName = String.format("<p><b>database</b>: %s</p>", formatDbName(props));
	
	return returnHeader + dbName + table + returnTrailer;  
	} // doGET_viewDBInfo


	/* doGET_viewLemmaInfoHTML:
	 * 
	 * GET /index/info/lemma/{lemma}?opts=o1+o2+o3...
	 * where: oi... = (flex)(+comp)(+other)(+spec) in any order.
	 * As opts is optional, flex is default.
	 * 
	 * E.g. /index/info/lemma/strahlen?opts=flex+other
	 * 
	 * Returns: text/html. 
	 * Returns wordforms for a submitted lemma. 
	 * 
	 * 13.01.20/FB
	 */
	
	@GET  
	@Path("info/lemma/{lemma}") 
	@Produces("text/html")
	 
	public String doGET_viewLemmaInfoHTML(
			@PathParam(value="lemma") 				  	final String lemma,
			@QueryParam("opts")  @DefaultValue("undef")	final String options,
			@QueryParam("prop1") @DefaultValue("3")		final int prop1,
			@QueryParam("prop2") @DefaultValue("3")		final int prop2,
			@QueryParam("spec")  @DefaultValue("3")		final int spec )
	 
	{
	final String
		func = "doGET_viewLemmaInfoHTML";
	int
		i, r;
	final String
		returnHeader = 	"<html><head>\n" + 
						" <meta charset=\"utf-8\">\n" + 
						"</head>\n" + 
						"<h1>GlemmServices: GET on /index/info/lemma</h1>\n" +
						"<h4>version " + version + " vom " + versionDate + " request = '" + LocalDateTime.now().toString().replace("T",  " ") + "'</i></h4>\n" +
						"<p><b>Query Syntax 1:</b> (C2 Options Setting)</p>" +
						"<p><code>{lemma}</code> <b>or</b> <code>{lemma}?opts=(flex)(+comp)(+other)(+spec)</code></p>\n" +
						"<p><b>Example:</b> <code>Öl</code> <b>or</b> <code>Öl?opts=flex+comp+other+spec</code></p>" +
						"<p><b>Query Syntax 2:</b> (internal Options Setting)</p>" +
						"<p><code>{lemma}?prop1=[012]&amp;prop2=[012]&amp;spec=[012]</code> (where each prop1, prop2, spec is optional).</p>\n" +
						"<p><b>Example:</b> <code>Öl?prop1=0&amp;prop2=1&amp;spec=2</code></p>";
	final String
		returnTrailer = "</html>";
	final String
		backgroundcolor = "style=\"background-color: #E0E0E0\"";
	StringBuilder
		table = new StringBuilder("<table border=\"1\" cellpadding=\"5\" cellspacing=\"0\">\n");
	String
		msg = null;
	int
		nErrors = 0,
		acceptedFormat = 0;
	LemmaInfo
		lemmaInfo = new LemmaInfo(lemma);
	Properties
		props;

	props = init_and_log_Request(func, "/index/info/lemma/{lemma}");
	
	fout.printf("Debug: %s: lemma='%s' options='%s' prop1='%d' prop2='%d' spec='%d'.\n", func, lemma, options, prop1, prop2, spec);
	
	/* 
	for(i=0; i<lemma.length(); i++)
		{
		fout.printf("Debug: %s: char at %d = '%c' = [%d].\n", func, i, lemma.charAt(i), lemma.codePointAt(i));
		}	
	 */
	
	nErrors = lemmaInfo.parseAllOptions(options, prop1, prop2, spec);
	
	if( nErrors != 0 )
		{
		ferr.printf("Error: %s: Errors in options Parameter!\n",  func); 
		msg = String.format("<p><b>Error:</b> options Parameter contains unknown values!</p>\n"); 
		return returnHeader + msg + returnTrailer;
		}
		
	acceptedFormat = getAcceptedFormat();
	
	table.append("<thead " + backgroundcolor + ">\n");
	if( lemmaInfo.optsSetting == LemmaInfo.OPTS_SETTING_C2 ) 
		msg = String.format("<tr><th>Lemma = '%s' &amp; options = <i>'%s'</i></th></tr>\n", lemmaInfo.lemmaOrig, lemmaInfo.optionstoString());
	else
		msg = String.format("<tr><th>Lemma = '%s' &amp; prop1,prop2,spec set directly.</th></tr>\n", lemmaInfo.lemmaOrig);

	table.append(msg);
	
	msg = String.format("<tr><th><b>prop1 = <i>'%s'</i> &amp; prop2 = <i>'%s'</i> &amp; spec = <i>'%s'</i></th></tr>\n",
			lemmaInfo.prop1 == 0 ? "-" : lemmaInfo.prop1 == 1 ? "+" : "undef", 
			lemmaInfo.prop2 == 0 ? "-" : lemmaInfo.prop2 == 1 ? "+" : "undef",
			lemmaInfo.spec  == 0 ? "-" : lemmaInfo.spec  == 1 ? "+" : "undef");
	table.append(msg);

	// extract lemmata from GlemmDB:
	GlemmDB.readDBLemmaData(lemmaInfo, props);
	
	// scan all wordforms and format them in the response buffer; 
	// if an error occures, wordform will be null;
	String 
		wordforms = lemmaInfo.wfs2HTML();
	
	// close sql stuff:
	int
		nWfs = lemmaInfo.nWfs;
	
	if( lemmaInfo != null )
		lemmaInfo.close();
	
	if( lemmaInfo.nWfs == 0 )
		wordforms = "<tr><td>keine Wortformen gefunden!</td></tr>\n";
	else if( lemmaInfo.nWfs < 0 || wordforms == null )
		wordforms = "<tr><td>Fehler aufgetreten: keine Wortformen gefunden!</td</tr>\n";
	else
		table.append("<tr><td><b>Anz. gefundene Wortformen</b> = " + lemmaInfo.nWfs + "</td></tr>\n");
		
	table.append("</thead>\n");
	
	// log outcome:
	if( nWfs < 0 )
		fout.printf("Error: %s: returning 0 results [HTML].\n", func);
	else
		fout.printf("Debug. %s: returning %d results [HTML].\n", func, nWfs);
	
	closeLogStreams();
	
	return returnHeader + table.toString() + "<tbody>\n" + wordforms + "</tbody>\n" + returnTrailer;
	
	} // viewLemmaInfoHTML
	
	/* viewLemmaInfoJSON:
	 * 
	 * GET /index/info/lemma/{lemma}?opts=o1+o2+o3...
	 * where: oi... = (flex)(+comp)(+other)(+spec) in any order.
	 * As opts is optional, flex is default.
	 * 
	 * E.g. /index/info/lemma/strahlen?opts=flex+other
	 * 
	 * Returns: application/json. 
	 * Returns wordforms for a submitted lemma. 
	 * 
	 * note:
	 * - if Browser request: HTTPServletRequest.accept = 'text/html,...xml...'
	 * - if WformServices request: HttpServletRequest.accept = 'application/json'.
	 * 
	 * 24.01.20/FB
	 * 
	 * - submitted URL, i.e. the requested lemma, is interpreted by glassfish as
	 *   being ISO-8859-1 encoded. When the request is UTF-8 encoded, Umlauts and
	 *   'ß' are mis-interpreded, leading to an erronerous request to the Derby
	 *   database.
	 * - Requested lemma 'lemma' has to be transformed to UTF-8 first.
	 * - note: even when submitting the GET request by specifying 'charencode=UTF-8',
	 *   glassfish does not convert it automatically as expected.
	 *   
	 *  22.10.21/FB
	 * 
	 * - if accept=application/json return UTF-8 (WformServices).
	 * - else if accept=text/html do not transform character encoding (Browser).
	 */
	
	@GET  
	@Path("info/lemma/{lemma}") 
	@Consumes("application/json, text/html")
	@Produces("application/json")
	
	public Response doGET_viewLemmaInfoJSON(
			@PathParam(value="lemma") 				 	final String lemma,
			@QueryParam("opts")  @DefaultValue("undef")	final String options,
			@QueryParam("prop1") @DefaultValue("3")		final int prop1,
			@QueryParam("prop2") @DefaultValue("3")		final int prop2,
			@QueryParam("spec")  @DefaultValue("3")		final int spec ) 
	{
	final String
		func = "doGET_viewLemmaInfoJSON";
	int
		nErrors = 0,
		acceptedFormat = 0;
	LemmaInfo
		lemmaInfo;
	LemmaResponse
		lemmaResp = null;
	Properties
		props;
	final boolean
		bEncodeLemma = false;
	
	props = init_and_log_Request(func, "/index/info/lemma/{lemma}?opts");
	
	/* test:
	for(int i=0; i<lemma.length(); i++)
	{
		fout.printf("Debug: %s: char at %d = '%c' = [%d].\n", func, i, lemma.charAt(i), lemma.codePointAt(i));
	}
	*/
	if( bEncodeLemma )
		{
		// when accepting JSON, url has to be converted to UTF-8, i.e. the requested lemma:
		try {
			lemmaInfo = new LemmaInfo(new String(lemma.getBytes("ISO-8859-1"), "utf-8"));
			} 
		catch (UnsupportedEncodingException e1) {
			e1.printStackTrace();
			LemmaResponse
				lemresp = new LemmaResponse();
			return Response.status(Response.Status.BAD_REQUEST).type(MediaType.APPLICATION_JSON).entity(lemresp).build();
			}
		fout.printf("Debug: %s: lemma='%s' [UTF-8] options='%s'.\n", func, lemmaInfo.lemmaOrig, options);
		}
	else
		{
		lemmaInfo = new LemmaInfo(lemma);
		fout.printf("Debug: %s: lemma='%s' [unchanged] options='%s'.\n", func, lemmaInfo.lemmaOrig, options);
		}

	/* does not help:
	 String
	 	requestCharEncoding = httpServletRequest.getCharacterEncoding();

	fout.printf("Debug: %s: request character encoding ='%s'.\n", func, requestCharEncoding);
	*/
	
	nErrors = lemmaInfo.parseAllOptions(options, prop1, prop2, spec);

	if( nErrors != 0 )
		{
		ferr.printf("Error: %s: Errors in options Parameter: '%s'!\n",  func, options); 
		LemmaResponse
			lemRespErr = new LemmaResponse();
		lemRespErr.head_errMess = String.format("Error: %s: options Parameter contains unknown values: '%s'!\n", func, options);
		
		return Response.status(Response.Status.BAD_REQUEST).type(MediaType.APPLICATION_JSON).entity(lemRespErr).build(); 
		}
	
	acceptedFormat = getAcceptedFormat();
	
	// extract lemmata from GlemmDB:
	GlemmDB.readDBLemmaData(lemmaInfo, props);
	
	// transform SQL Response to JSON Response:
	lemmaResp = lemmaInfo.res2JSON();
	
	fout.printf("Debug: %s: lemmaResp: nWfs=%d errMess='%s' list='%s'.\n", func, lemmaResp.head_nWfs, 
						lemmaResp.head_errMess != null ? lemmaResp.head_errMess : "-",
						lemmaResp.listofWfs);
	
	// close:
	if( lemmaInfo != null )
		lemmaInfo.close();
	
	closeLogStreams();
	
	/* return Lemma Response Object */
	if( lemmaResp.head_errMess == null )
		{
		return Response.status(Response.Status.OK).type(MediaType.APPLICATION_JSON).entity(lemmaResp).build(); 
		//return Response.status(Response.Status.OK).entity(lemmaResp).build(); 
		}
	else // on errors:
		return Response.status(Response.Status.INTERNAL_SERVER_ERROR).type(MediaType.APPLICATION_JSON).entity(lemmaResp).build();

	} // viewLemmaInfoJSON
	
	/* viewLemmaInfoTEXT:
	 * 
	 * either:
	 * a) GET /index/info/lemma/{lemma}?opts=(flex)(+comp)(+other)(+spec)
	 *    this is the original C2 Glemm Option Setting and should be totally
	 *    compatible to the way Glemm Options work in COSMAS II.
	 * b) GET /index/info/lemma/{lemma}?(prop1=n)(&prop2=n)(&spec=n)
	 *    this is the extended mode in KorAP which lets every flag from 
	 *    the Glemm Lemma table be set individually;
	 *    where n=0 : false, n=1 : true; n=2 : undef.
	 *    
	 * This Service expects first "opts" to be set (precedence other the 
	 * other options) -> C2 Option Setting, or 
	 * one of "prop1", "prop2" and "spec", in which case -> extended flag
	 * setting is used.
	 * 
	 * Default is ?opts=flex.
	 * 
	 * E.g. /index/info/lemma/strahlen?opts=flex+other
	 * E.g. /index/info/lemma/strahlen?prop1=0&prop2=1&spec=2
	 * 
	 * Notes:
	 *          
	 * Returns: text/plain. 
	 * Returns only wordforms for a submitted lemma. 
	 * 
	 * 31.01.20/FB
	 */
	
	@GET  
	@Path("info/lemma/{lemma}") 
	@Produces("text/plain")
	 
	public Response doGET_viewLemmaInfoTEXT(
			@PathParam(value="lemma") 				 	final String lemma,
			@QueryParam("opts")  @DefaultValue("undef")	final String options,
			@QueryParam("prop1") @DefaultValue("3")		final int prop1,
			@QueryParam("prop2") @DefaultValue("3")		final int prop2,
			@QueryParam("spec")  @DefaultValue("3")		final int spec )
	 
	{
	final String
		func = "doGET_viewLemmaInfoTEXT";
	int
		nErrors = 0,
		acceptedFormat = 0; // undef.
	String
		result = "";
	LemmaInfo
		lemmaInfo = new LemmaInfo(lemma);
	LemmaResponse
		lemmaResp = null;
	Properties
		props;

	props = init_and_log_Request(func, "/index/info/lemma/{lemma}?opts");	

	fout.printf("Debug: %s: lemma='%s' options='%s' prop1=%d prop2=%d spec=%d.\n", 
			func, lemma, options, prop1, prop2, spec);

	/*
	for(int i=0; i<lemma.length(); i++)
	{
		fout.printf("Debug: %s: char at %d = '%c' = [%d].\n", func, i, lemma.charAt(i), lemma.codePointAt(i));
	}
	*/
	acceptedFormat = getAcceptedFormat();
		
	nErrors = lemmaInfo.parseAllOptions(options, prop1, prop2, spec);
	
	if( nErrors != 0 )
		{
		ferr.printf("Error: %s: Errors in options Parameter: '%s'!\n",  func, options); 
		LemmaResponse
			lemRespErr = new LemmaResponse();
		lemRespErr.head_errMess = String.format("Error: %s: options Parameter contains unknown values: '%s'!\n", func, options);
		
		return Response.status(Response.Status.BAD_REQUEST).type(MediaType.APPLICATION_JSON).entity(lemRespErr).build(); 
		}
		
	
	// extract lemmata from GlemmDB:
	GlemmDB.readDBLemmaData(lemmaInfo, props);
	
	// transform SQL Response to JSON Response:
	switch(acceptedFormat)
		{
	case ACC_FORMAT_TEXT:
		result = lemmaInfo.res2TEXT();
		break;
	case ACC_FORMAT_JSON:
		// transform SQL Response to JSON Response:
		lemmaResp = lemmaInfo.res2JSON();
		break;
	case ACC_FORMAT_HTML:
		// not implemented here, see viewLemmaInfoHTML().
		result = null;
		break;
	default:
		result = lemmaInfo.res2TEXT();
		}
	
	// close:
	if( lemmaInfo != null )
		lemmaInfo.close();
	
	closeLogStreams();
	
	switch( acceptedFormat)
		{
	case ACC_FORMAT_TEXT:
		/* return Lemma Response Object */
		if( result != null )
			return Response.status(Response.Status.ACCEPTED).type(MediaType.TEXT_PLAIN).entity(result).build(); 
		else// on errors:
			return Response.status(Response.Status.INTERNAL_SERVER_ERROR).type(MediaType.TEXT_PLAIN).entity("").build();
		// break;
	
	case ACC_FORMAT_JSON:
		/* return Lemma Response Object */
		if( lemmaResp.head_errMess != null )
			return Response.status(Response.Status.ACCEPTED).type(MediaType.APPLICATION_JSON).entity(lemmaResp).build(); 
		else// on errors:
			return Response.status(Response.Status.INTERNAL_SERVER_ERROR)
						.type(MediaType.APPLICATION_JSON).entity(lemmaResp).build();
		// break;

	case ACC_FORMAT_HTML:
		return Response.status(Response.Status.INTERNAL_SERVER_ERROR)
				.type(MediaType.APPLICATION_XML_TYPE).entity("internal error: not implemented in this procedure!").build();
	default:
		return Response.status(Response.Status.INTERNAL_SERVER_ERROR)
				.type(MediaType.APPLICATION_XML_TYPE).entity("internal error: requested Format is not known!").build();
		}


	} // viewLemmaInfoTEXT
	
	/*
	 * Test function: 
	 * Returns current content of Lemma Database.
	 * 09.10.20/FB
	 */
	@GET @Path("test/info") @Produces("text/html")
	 
	public String viewDbContent() 
	 
	{
	final String
		func = "viewDbContent";
	int
		r;
	String
		returnHeader = "<html><h1>GlemmServices: GET on /index/test/info</h1>" +
					   "<p>Version vom " + versionDate + "</p>";
	String
		returnTrailer = "</html>";
	String
		msg;
	GlemmDBInfo
		dbInfo;
	Properties
		props;

	props = init_and_log_Request(func, "/index/test/info/");

	dbInfo = GlemmDB.readDbInfos(props);
	if( dbInfo == null )
		msg = "<p><b>Fehler:</b> Inhalt der Datebank kann nicht gelesen werden!</p>";
	else
		msg = "<p>Alles gut.</p>";

	closeLogStreams();
	
	return returnHeader + msg + returnTrailer;  
	}
	
	/*
	 * getAcceptedFormat
	 * 17.02.20/FB
	 */
	 
	 public int getAcceptedFormat()
	 
	 {
	 final String
	 	func = "getAcceptedFormat";
	Enumeration<String>
		names = httpServletRequest.getHeaderNames();
	 
	//fout.printf("Debug: %s: HTTP Request:\n", func);

	while( names.hasMoreElements() )
		{
		String
			key = (String)names.nextElement(),
			val = (String)httpServletRequest.getHeader(key);
		
		fout.printf("Header: '%s' = '%s'.\n", key, val);
		if( key.equals("accept") )
			{
			switch( val )
				{
			case "text/plain": 
				fout.printf("accepted format = 'text/plain': ok.\n");
				return ACC_FORMAT_TEXT;
			case "application/json": 
				fout.printf("accepted format = 'applic/JSON': ok.\n");
				return ACC_FORMAT_JSON;
			case "text/html": 
				fout.printf("accepted format = 'applic/HTML': ok.\n");
				return ACC_FORMAT_HTML;
			default: 
				fout.printf("accepted format = '%s' = unknown! Default='text/plain'.\n", val);
				return ACC_FORMAT_TEXT;
				}
			}
		}

	return 0;	// something wrong.	

 	} // getAcceptedFormat	

}
